# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#


class PerChannel(object):
    """
    Utility class, usable with Process() and Call(), that delegates events
    to different instances of the same class, using one instance per
    channel/port combination.
    New instances are created as needed, using the given factory function,
    when the first event with a new channel/port arrives.
    """
    def __init__(self, factory):
        self.channel_proc = {}
        self.factory = factory

    def __call__(self, ev):
        k = (ev.port, ev.channel)
        if k not in self.channel_proc:
            self.channel_proc[k] = self.factory()
        return self.channel_proc[k](ev)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

import mididings.constants as _constants
import mididings.arguments as _arguments
import mididings.util as _util
import mididings.misc as _misc



def _make_property(type, data, name=None, offset=False):
    if isinstance(data, str) and not offset:
        def getter(self):
            self._check_type_attribute(type, name)
            return getattr(self, data)
        def setter(self, value):
            self._check_type_attribute(type, name)
            setattr(self, data, value)

    elif isinstance(data, str) and offset:
        def getter(self):
            self._check_type_attribute(type, name)
            return _util.offset(getattr(self, data))
        def setter(self, value):
            self._check_type_attribute(type, name)
            setattr(self, data, _util.actual(value))

    return property(getter, setter)


class MidiEvent(_mididings.MidiEvent):
    """
    MidiEvent(type, port=0, channel=0, data1=0, data2=0, sysex=None)

    A MIDI event, as seen by Python code.
    """
    @_arguments.accept(
        None, _constants._EventType,
        _util.port_number, _util.channel_number,
        int, int, _arguments.nullable(_util.sysex_data)
    )
    def __init__(
        self, type,
        port=_util.NoDataOffset(0), channel=_util.NoDataOffset(0),
        data1=0, data2=0, sysex=None
    ):
        # use default c'tor, then set data members manually to take advantage
        # of automatic data offset conversion
        _mididings.MidiEvent.__init__(self)
        self.type = type
        self.port = port
        self.channel = channel
        self.data1 = data1
        self.data2 = data2
        if sysex is not None:
            self.sysex_ = sysex

    # mask boost python safety-guard, as we don't use __getstate__, but __getinitargs__
    # more info: https://beta.boost.org/doc/libs/1_53_0/libs/python/doc/v2/pickle.html
    __getstate_manages_dict__ = 1

    def __getinitargs__(self):
        self._finalize()
        return (self.type, self.port, self.channel, self.data1, self.data2,
                self.sysex if self.type == _constants.SYSEX else None)

    def _check_type_attribute(self, type, name):
        if not self.type & type:
            message = (f"MidiEvent type '{self._type_to_string()}' has no attribute '{name}'")
            raise AttributeError(message)

    def _type_to_string(self):
        try:
            return _constants._EVENT_TYPES[self.type].name
        except KeyError:
            return 'NONE'

    def _sysex_to_hex(self, max_length):
        data = self.sysex
        if max_length:
            m = max(0, max_length // 3)
            if len(data) > m:
                return f"{_misc.sequence_to_hex(data[:m])} ..."
        return _misc.sequence_to_hex(data)

    _to_string_mapping = {
        _constants.NOTEON:
            lambda self: 'Note On:  %3d %3d  (%s)' %
                        (self.note, self.velocity, _util.note_name(self.note)),
        _constants.NOTEOFF:
            lambda self: 'Note Off: %3d %3d  (%s)' %
                        (self.note, self.velocity, _util.note_name(self.note)),
        _constants.CTRL:
            lambda self: 'Ctrl:     %3d %3d' % (self.ctrl, self.value) +
                            ('  (%s)' % _util.controller_name(self.ctrl)
                                if _util.controller_name(self.ctrl) else ''),
        _constants.PITCHBEND:
            lambda self: 'Pitchbend:  %5d' % self.value,
        _constants.AFTERTOUCH:
            lambda self: 'Aftertouch:   %3d' % self.value,
        _constants.POLY_AFTERTOUCH:
            lambda self: 'Poly Aftertouch: %3d %3d  (%s)' %
                        (self.note, self.value, _util.note_name(self.note)),
        _constants.PROGRAM:
            lambda self: 'Program:      %3d' % self.program,
        _constants.SYSCM_QFRAME:
            lambda self: 'SysCm QFrame: %3d' % self.data1,
        _constants.SYSCM_SONGPOS:
            lambda self: 'SysCm SongPos:%3d %3d' % (self.data1, self.data2),
        _constants.SYSCM_SONGSEL:
            lambda self: 'SysCm SongSel:%3d' % self.data1,
        _constants.SYSCM_TUNEREQ:
            lambda self: 'SysCm TuneReq',
        _constants.SYSRT_CLOCK:
            lambda self: 'SysRt Clock',
        _constants.SYSRT_START:
            lambda self: 'SysRt Start',
        _constants.SYSRT_CONTINUE:
            lambda self: 'SysRt Continue',
        _constants.SYSRT_STOP:
            lambda self: 'SysRt Stop',
        _constants.SYSRT_SENSING:
            lambda self: 'SysRt Sensing',
        _constants.SYSRT_RESET:
            lambda self: 'SysRt Reset',
        _constants.DUMMY:
            lambda self: 'Dummy',
    }

    _repr_mapping = {
        _constants.NOTEON:
            lambda self:
                'NoteOnEvent(port=%d, channel=%d, note=%d, velocity=%d)' %
                    (self.port, self.channel, self.note, self.velocity),
        _constants.NOTEOFF:
            lambda self:
                'NoteOffEvent(port=%d, channel=%d, note=%d, velocity=%d)' %
                    (self.port, self.channel, self.note, self.velocity),
        _constants.CTRL:
            lambda self:
                'CtrlEvent(port=%d, channel=%d, ctrl=%d, value=%d)' %
                    (self.port, self.channel, self.ctrl, self.value),
        _constants.PITCHBEND:
            lambda self:
                'PitchbendEvent(port=%d, channel=%d, value=%d)' %
                    (self.port, self.channel, self.value),
        _constants.AFTERTOUCH:
            lambda self:
                'AftertouchEvent(port=%d, channel=%d, value=%d)' %
                    (self.port, self.channel, self.value),
        _constants.PROGRAM:
            lambda self:
                'ProgramEvent(port=%d, channel=%d, program=%d)' %
                    (self.port, self.channel, self.program),
        _constants.SYSEX:
            lambda self:
                'SysExEvent(port=%d, sysex=%r)' %
                    (self.port, _misc.bytestring(self._get_sysex())),
    }

    def to_string(self, portnames=[], portname_length=0, max_length=0):
        if len(portnames) > self.port_:
            port = portnames[self.port_]
        else:
            port = str(self.port)

        header = '[%*s, %2d]' % (max(portname_length, 2), port, self.channel)

        if self.type_ == _constants.SYSEX:
            maxsysex = (max_length - len(header) - 25) if max_length else 0
            sysexstr = self._sysex_to_hex(maxsysex)
            desc = 'SysEx:   %8d  [%s]' % (len(self.sysex), sysexstr)
        else:
            desc = MidiEvent._to_string_mapping.get(
                self.type_,
                lambda self: 'None'
            )(self)

        return '%s %s' % (header, desc)

    def __repr__(self):
        return MidiEvent._repr_mapping.get(
            self.type_,
            # default for types not in _repr_mapping
            lambda self:
                'MidiEvent(%s, %d, %d, %d, %d)' %
                    (self._type_to_string(), self.port, self.channel,
                    self.data1, self.data2)
        )(self)


    def __eq__(self, other):
        if not isinstance(other, MidiEvent):
            return NotImplemented
        self._finalize()
        other._finalize()
        return _mididings.MidiEvent.__eq__(self, other)

    def __ne__(self, other):
        if not isinstance(other, MidiEvent):
            return NotImplemented
        self._finalize()
        other._finalize()
        return _mididings.MidiEvent.__ne__(self, other)

    def __hash__(self):
        self._finalize()
        return _mididings.MidiEvent.__hash__(self)

    def _finalize(self):
        if hasattr(self, '_sysex_tmp'):
            self.sysex_ = _util.sysex_data(self._sysex_tmp)
            delattr(self, '_sysex_tmp')


    def _type_getter(self):
        # return an event type constant, rather than the plain int stored in
        # self.type_
        return _constants._EVENT_TYPES[self.type_]

    def _type_setter(self, type):
        self.type_ = type

    def _get_sysex(self):
        if hasattr(self, '_sysex_tmp'):
            return self._sysex_tmp
        else:
            return _util.sysex_to_bytearray(self.sysex_)

    def _sysex_getter(self):
        self._check_type_attribute(_constants.SYSEX, 'sysex')
        self._sysex_tmp = self._get_sysex()
        return self._sysex_tmp

    def _sysex_setter(self, sysex):
        self._check_type_attribute(_constants.SYSEX, 'sysex')
        self._sysex_tmp = _util.sysex_to_bytearray(sysex)


    #: The event type, one of the :ref:`event-types` constants.
    type = property(_type_getter, _type_setter)

    #: The port number.
    port      = _make_property(_constants.ANY, 'port_', offset=True)
    #: The channel number.
    channel   = _make_property(_constants.ANY, 'channel_', offset=True)

    #: The note number, stored in :attr:`data1`.
    note      = _make_property(_constants.NOTE |
                               _constants.POLY_AFTERTOUCH,
                               'data1', 'note')

    #: The velocity value, stored in :attr:`data2`.
    velocity  = _make_property(_constants.NOTE, 'data2', 'velocity')

    #: The controller number, stored in :attr:`data1`.
    ctrl      = _make_property(_constants.CTRL, 'data1', 'ctrl')

    #: The controller value, stored in :attr:`data2`.
    value     = _make_property(_constants.CTRL | _constants.PITCHBEND |
                               _constants.AFTERTOUCH |
                               _constants.POLY_AFTERTOUCH,
                               'data2', 'value')

    #: The program number, stored in :attr:`data2`.
    #: Unlike :attr:`data2`, this attribute observes the
    #: :c:data:`data_offset` setting.
    program   = _make_property(_constants.PROGRAM,
                               'data2', 'program', offset=True)

    #: SysEx data.
    sysex = property(_sysex_getter, _sysex_setter)



@_arguments.accept(_util.port_number, _util.channel_number,
                   _util.note_number, int)
def NoteOnEvent(port, channel, note, velocity):
    """
    Create a new note-on event object.
    """
    return MidiEvent(_constants.NOTEON, port, channel, note, velocity)

@_arguments.accept(_util.port_number, _util.channel_number,
                   _util.note_number, int)
def NoteOffEvent(port, channel, note, velocity=0):
    """
    Create a new note-off event object.
    """
    return MidiEvent(_constants.NOTEOFF, port, channel, note, velocity)

@_arguments.accept(_util.port_number, _util.channel_number,
                   _util.ctrl_number, int)
def CtrlEvent(port, channel, ctrl, value):
    """
    Create a new control change event object.
    """
    return MidiEvent(_constants.CTRL, port, channel, ctrl, value)

@_arguments.accept(_util.port_number, _util.channel_number, int)
def PitchbendEvent(port, channel, value):
    """
    Create a new pitch bend event object.
    """
    return MidiEvent(_constants.PITCHBEND, port, channel, 0, value)

@_arguments.accept(_util.port_number, _util.channel_number, int)
def AftertouchEvent(port, channel, value):
    """
    Create a new aftertouch event object.
    """
    return MidiEvent(_constants.AFTERTOUCH, port, channel, 0, value)

@_arguments.accept(_util.port_number, _util.channel_number,
                   _util.note_number, int)
def PolyAftertouchEvent(port, channel, note, value):
    """
    Create a new polyphonic aftertouch event object.
    """
    return MidiEvent(_constants.POLY_AFTERTOUCH, port, channel, note, value)

@_arguments.accept(_util.port_number, _util.channel_number,
                   _util.program_number)
def ProgramEvent(port, channel, program):
    """
    Create a new program change event object.
    """
    return MidiEvent(_constants.PROGRAM, port, channel, 0,
                     _util.actual(program))

@_arguments.accept(_util.port_number, _util.sysex_data)
def SysExEvent(port, sysex):
    """
    Create a new sysex event object.
    """
    return MidiEvent(_constants.SYSEX, port, sysex=sysex)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import mididings as _m
import mididings.event as _event
import mididings.util as _util
from mididings.extra.per_channel import PerChannel as _PerChannel


class _LatchNotes(object):
    def __init__(self, polyphonic, reset):
        self.polyphonic = polyphonic
        self.reset = _util.note_number(reset) if reset is not None else None
        self.notes = []

    def __call__(self, ev):
        if ev.type == _m.NOTEON:
            if ev.note == self.reset:
                # reset all notes
                r = [_event.NoteOffEvent(ev.port, ev.channel, x, 0)
                        for x in self.notes]
                self.notes = []
                return r

            if self.polyphonic:
                if ev.note not in self.notes:
                    # turn note on
                    self.notes.append(ev.note)
                    return ev
                else:
                    # turn note off
                    self.notes.remove(ev.note)
                    return _event.NoteOffEvent(ev.port, ev.channel, ev.note, 0)
            else:
                # monophonic: turn off previous note, play new note
                if len(self.notes):
                    r = [_event.NoteOffEvent(
                                    ev.port, ev.channel, self.notes[0], 0)]
                else:
                    r = []
                self.notes = [ev.note]
                return r + [ev]

        elif ev.type == _m.NOTEOFF:
            # ignore all note-off events
            return None


def LatchNotes(polyphonic=False, reset=None):
    """
    Makes notes latching, so they will keep playing when the key is
    released.

    :param polyphonic:
        If true, each note can be stopped individually by pressing the
        corresponding key again.
        Otherwise pressing a key will automatically turn off any previous
        notes.

    :param reset:
        a note (name/number) that acts as a reset key, stopping all
        currently playing notes.
    """
    return (_m.Filter(_m.NOTE) %
                _m.Process(_PerChannel(
                    lambda: _LatchNotes(polyphonic, reset))))



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

import mididings.patch as _patch
import mididings.scene as _scene
import mididings.util as _util
import mididings.misc as _misc
import mididings.setup as _setup
import mididings.constants as _constants
import mididings.overload as _overload
import mididings.arguments as _arguments
from mididings.units.base import _UNIT_TYPES
from mididings.scene import _SCENE_TYPES

import time as _time
import weakref as _weakref
import threading as _threading
import gc as _gc
import atexit as _atexit
import os as _os
import sys as _sys

_TheBackend = None
_TheEngine = None


def _start_backend():
    global _TheBackend
    if _TheBackend is None:
        _TheBackend = _mididings.create_backend(
            _setup.get_config('backend'),
            _setup.get_config('client_name'),
            _setup._in_portnames,
            _setup._out_portnames
        )
        if _TheBackend:
            _TheBackend.connect_ports(_setup._in_port_connections,
                                      _setup._out_port_connections)


class Engine(_mididings.Engine):
    def __init__(self):
        _start_backend()

        verbose = not _setup.get_config('silent')
        # initialize C++ base class
        _mididings.Engine.__init__(self, _TheBackend, verbose)

        self._scenes = {}

    def _build_patches(self, sceneobj, number):
        patch = _patch.Patch(sceneobj.patch)
        init_patch = _patch.Patch(sceneobj.init_patch)
        exit_patch = _patch.Patch(sceneobj.exit_patch)
        # add scene to base class object
        self.add_scene(_util.actual(number), patch, init_patch, exit_patch)

    def setup(self, scenes, control, pre, post):
        # build and setup all scenes and scene groups
        for number, scene in scenes.items():
            if isinstance(scene, _scene.SceneGroup):
                self._scenes[number] = (scene.name, [])

                for subscene in scene.subscenes:
                    sceneobj = _scene._parse_scene(subscene)
                    self._scenes[number][1].append(sceneobj.name)
                    self._build_patches(sceneobj, number)
            else:
                sceneobj = _scene._parse_scene(scene)
                self._scenes[number] = (sceneobj.name, [])
                self._build_patches(sceneobj, number)

        # build and setup control, pre, and post patches
        control_patch = _patch.Patch(control) if control else None
        pre_patch = _patch.Patch(pre) if pre else None
        post_patch = _patch.Patch(post) if post else None
        # tell base class object about these patches
        self.set_processing(control_patch, pre_patch, post_patch)

        global _TheEngine
        _TheEngine = _weakref.ref(self)

        _gc.collect()
        _gc.disable()

    def run(self):
        self._quit = _threading.Event()

        # delay before actually sending any midi data (give qjackctl
        # patchbay time to react...)
        self._start_delay()

        self._call_hooks('on_start')

        initial_scene, initial_subscene = \
            self._parse_scene_number(_setup.get_config('initial_scene'))

        # start the actual event processing
        self.start(initial_scene, initial_subscene)

        try:
            # wait() with no timeout also blocks KeyboardInterrupt, but
            # a very long timeout doesn't. weird...
            while not self._quit.isSet():
                self._quit.wait(86400)
        except KeyboardInterrupt:
            pass
        finally:
            self._call_hooks('on_exit')
            global _TheEngine
            _TheEngine = None

    def _start_delay(self):
        delay = _setup.get_config('start_delay')
        if delay is not None:
            if delay > 0:
                _time.sleep(delay)
            else:
                input("press enter to start midi processing...")

    def _parse_scene_number(self, number):
        if number in self._scenes:
            # single scene number, no subscene
            return (_util.actual(number), -1)
        elif (_misc.issequence(number) and
                len(number) > 1 and number[0] in self._scenes):
            # scene/subscene numbers as tuple...
            if _util.actual(number[1]) < len(self._scenes[number[0]][1]):
                # both scene and subscene numbers are valid
                return (_util.actual(number[0]), _util.actual(number[1]))
            # subscene number is invalid
            return (_util.actual(number[0]), -1)
        # no such scene
        return (-1, -1)

    def scene_switch_callback(self, scene, subscene):
        # scene and subscene parameters are the actual numbers without offset!
        if scene == -1:
            # no scene specified, use current
            scene = _mididings.Engine.current_scene(self)
        if subscene == -1:
            # no subscene specified, use first
            subscene = 0

        # save actual subscene index
        subscene_index = subscene

        # add data offset to scene/subscene numbers
        scene = _util.offset(scene)
        subscene = _util.offset(subscene)

        found = (scene in self._scenes and
                 (not subscene_index or
                    subscene_index < len(self._scenes[scene][1])))

        # get string representation of scene/subscene number
        if (subscene_index or
                (scene in self._scenes and len(self._scenes[scene][1]))):
            number = f"{int(scene)}.{int(subscene)}"
        else:
            number = str(scene)

        if not _setup.get_config('silent'):
            if found:
                # get scene/subscene name
                scene_data = self._scenes[scene]
                if scene_data[1]:
                    name = f"{scene_data[0]} - {scene_data[1][subscene_index]}"
                else:
                    name = scene_data[0]

                scene_desc = (f"{number}: {name}" if name else str(number))
                print(f"switching to scene {scene_desc}")
            else:
                print(f"no such scene: {number}")

        if found:
            self._call_hooks('on_switch_scene', scene, subscene)

    def _call_hooks(self, name, *args):
        for hook in _setup.get_hooks():
            if hasattr(hook, name):
                f = getattr(hook, name)
                f(*args)

    def switch_scene(self, scene, subscene=None):
        _mididings.Engine.switch_scene(self,
            _util.actual(scene),
            _util.actual(subscene) if subscene is not None else -1
         )

    def switch_subscene(self, subscene):
        _mididings.Engine.switch_scene(self, -1, _util.actual(subscene))

    def current_scene(self):
        return _util.offset(_mididings.Engine.current_scene(self))

    def current_subscene(self):
        return _util.offset(_mididings.Engine.current_subscene(self))

    def scenes(self):
        return self._scenes

    def process_event(self, ev):
        ev._finalize()
        return _mididings.Engine.process_event(self, ev)

    def output_event(self, ev):
        ev._finalize()
        _mididings.Engine.output_event(self, ev)

    def process(self, ev):
        ev._finalize()
        return _mididings.Engine.process(self, ev)

    def restart(self):
        _atexit.register(self._restart)
        self.quit()

    @staticmethod
    def _restart():
        # run the same interpreter with the same arguments again
        _os.execl(_sys.executable, _sys.executable, *_sys.argv)

    def quit(self):
        self._quit.set()



@_overload.mark(
    """
    run(patch)
    run(scenes=..., control=None, pre=None, post=None)

    Start the MIDI processing. This is usually the last function called by a
    mididings script.
    The first version just runs a single patch, while the second version
    allows switching between multiple scenes.

    :param patch: a single patch.
    :param scenes: a dictionary with program numbers as keys, and
        :class:`Scene` objects, :class:`SceneGroup` objects or plain patches
        as values.
    :param control: an optional "control" patch, which is always active, and
        runs in parallel to the current scene.
    :param pre: an optional patch that allows common processing to take place
        before every scene. Does not affect the control patch.
    :param post: an optional patch that allows common processing to take place
        after every scene. Does not affect the control patch.
    """
)
@_arguments.accept(_UNIT_TYPES)
def run(patch):
    if (isinstance(patch, dict) and
            all(not isinstance(k, _constants._EventType)
                for k in patch.keys())):
        # bypass the overload mechanism (just this once...) if there's no way
        # the given dict could be accepted as a split
        run(scenes=patch)
    else:
        e = Engine()
        e.setup({_util.offset(0): patch}, None, None, None)
        e.run()

@_overload.mark
@_arguments.accept(
    {_util.scene_number: _SCENE_TYPES},
    _arguments.nullable(_UNIT_TYPES),
    _arguments.nullable(_UNIT_TYPES),
    _arguments.nullable(_UNIT_TYPES)
)
def run(scenes, control=None, pre=None, post=None):
    e = Engine()
    e.setup(scenes, control, pre, post)
    e.run()


def switch_scene(scene, subscene=None):
    """
    Switch to the given scene number.
    """
    _TheEngine().switch_scene(scene, subscene)

def switch_subscene(subscene):
    """
    Switch to the given subscene number.
    """
    _TheEngine().switch_subscene(subscene)

def current_scene():
    """
    Return the current scene number.
    """
    return _TheEngine().current_scene()

def current_subscene():
    """
    Return the current subscene number.
    """
    return _TheEngine().current_subscene()

def scenes():
    """
    Return a dictionary of all scenes. Keys are scene numbers, values are
    tuples containing the scene name and a (possibly empty) list of subscene
    names.
    """
    return _TheEngine().scenes()

def output_event(ev):
    """
    Send an event directly to an output port, completely bypassing any other
    event processing.
    """
    _TheEngine().output_event(ev)

def in_ports():
    """
    Return a list of the configured input port names.
    """
    return _setup._in_portnames

def out_ports():
    """
    Return a list of the configured output port names.
    """
    return _setup._out_portnames

def time():
    """
    Return the time in seconds (floating point) since some unspecified
    starting point. Unlike Python's :func:`time.time()`, this clock is
    guaranteed to be monotonic.
    """
    return _TheEngine().time()

def get_client_name():
    """
    Return actual client name
    """
    return _TheEngine().get_client_name()

def get_client_id():
    """
    Return client id
    """
    if _setup.get_config('backend') == 'alsa':
        return _TheEngine().get_client_id()
    uuid = _TheEngine().get_client_uuid()
    if uuid.isdigit():
        return int(uuid)
    return uuid

def active():
    """
    Return ``True`` if the mididings engine is active (the :func:`~.run()`
    function is running).
    """
    return _TheEngine is not None and _TheEngine() is not None

def restart():
    """
    Restart the mididings script by terminating the current process, and then
    running the same Python interpreter with the same arguments again.
    This will not work properly if :func:`~.run()` is not the last call in
    your script, or if you're running mididings in an interactive Python
    interpreter.
    """
    _TheEngine().restart()

def quit():
    """
    Terminate mididings by making the :func:`~.run()` function return.
    """
    _TheEngine().quit()



def process_file(infile, outfile, patch):
    """
    Process a MIDI file, using pysmf.
    """
    import smf

    # create dummy engine with no inputs or outputs
    _setup._config_impl(backend='dummy')
    engine = Engine()
    engine.setup({_util.offset(0): patch}, None, None, None)

    # open input file
    smf_in = smf.SMF(infile)

    # create SMF object with same ppqn and number of tracks as input file
    smf_out = smf.SMF()
    smf_out.ppqn = smf_in.ppqn
    for n in range(len(smf_in.tracks)):
        smf_out.add_track()

    for smf_ev in smf_in.events:
        if smf_ev.midi_buffer[0] == 0xff:
            # event is metadata. copy to output unmodified
            smf_out.add_event(smf.Event(smf_ev.midi_buffer),
                              smf_ev.track_number, pulses=smf_ev.time_pulses)
        else:
            ev = _mididings.buffer_to_midi_event(smf_ev.midi_buffer,
                              smf_ev.track_number, smf_ev.time_pulses)

            # use base class version of process_event() to bypass calling
            # ev._finalize(), which would fail since ev is of type
            # _mididings.MidiEvent.
            for out_ev in _mididings.Engine.process_event(engine, ev):
                buf, track, pulses = _mididings.midi_event_to_buffer(out_ev)
                smf_out.add_event(smf.Event(buf), track, pulses=pulses)

    smf_out.save(outfile)



# -*- coding: utf-8 -*-
#
# mididings
#
# Copyright (C) 2008-2014  Dominic Sacré  <dominic.sacre@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#

import _mididings

import mididings.constants as _constants
import mididings.arguments as _arguments
import mididings.unitrepr as _unitrepr


class _Unit(object):
    """
    Wrapper class for all units.
    """
    def __init__(self, unit):
        self.unit = unit

    def __rshift__(self, other):
        """
        Connect units in series (operator >>).
        """
        if not isinstance(other, _UNIT_TYPES):
            return NotImplemented
        return _join_units(_Chain, self, other)

    def __rrshift__(self, other):
        """
        Connect units in series (operator >>).
        """
        if not isinstance(other, _UNIT_TYPES):
            return NotImplemented
        return _join_units(_Chain, other, self)

    def __floordiv__(self, other):
        """
        Connect units in parallel (operator //).
        """
        if not isinstance(other, _UNIT_TYPES):
            return NotImplemented
        return _join_units(_Fork, self, other)

    def __rfloordiv__(self, other):
        """
        Connect units in parallel (operator //).
        """
        if not isinstance(other, _UNIT_TYPES):
            return NotImplemented
        return _join_units(_Fork, other, self)

    def __pos__(self):
        """
        Apply to duplicate (unary operator +)
        """
        return self.add()

    def add(self):
        return Fork([ Pass(), self ])

    def __repr__(self):
        return _unitrepr.unit_to_string(self)


def _join_units(t, a, b):
    """
    Combine units in a single instance of type t, avoiding nesting if possible.
    """
    if not isinstance(a, t):
        a = [a]
    if not isinstance(b, t):
        b = [b]
    return t(a + b)


class _Chain(_Unit, list):
    def __init__(self, units):
        list.__init__(self, units)

    def __repr__(self):
        return _unitrepr.chain_to_string(self)


class _Fork(_Unit, list):
    def __init__(self, units, remove_duplicates=None):
        list.__init__(self, units)
        self.remove_duplicates = remove_duplicates

    def __repr__(self):
        return _unitrepr.fork_to_string(self)


class _Split(_Unit, dict):
    def __init__(self, mapping):
        dict.__init__(self, mapping)

    def __repr__(self):
        return _unitrepr.split_to_string(self)


class _Selector(object):
    """
    Base class for anything that can act as a selector.
    Derived classes must implement methods build() and build_negated().
    """
    def __and__(self, other):
        """
        Return conjunction of multiple filters (operator &).
        """
        if not isinstance(other, _Selector):
            return NotImplemented
        return _join_selectors(_AndSelector, self, other)

    def __or__(self, other):
        """
        Return disjunction of multiple filters (operator |).
        """
        if not isinstance(other, _Selector):
            return NotImplemented
        return _join_selectors(_OrSelector, self, other)

    def __mod__(self, other):
        """
        Apply the selector (operator %).
        """
        if isinstance(other, tuple) and len(other) in (1, 2):
            return self.apply(*other)
        else:
            return self.apply(other)

    def apply(self, patch, else_patch=None):
        if else_patch is None:
            return Fork([
                self.build() >> patch,
                self.build_negated(),
            ])
        else:
            return Fork([
                self.build() >> patch,
                self.build_negated() >> else_patch,
            ])


class _AndSelector(_Selector):
    def __init__(self, conditions):
        self.conditions = conditions

    def build(self):
        return Chain(p.build() for p in self.conditions)

    def build_negated(self):
        return Fork(p.build_negated() for p in self.conditions)


class _OrSelector(_Selector):
    def __init__(self, conditions):
        self.conditions = conditions

    def build(self):
        return Fork(p.build() for p in self.conditions)

    def build_negated(self):
        return Chain(p.build_negated() for p in self.conditions)


def _join_selectors(t, a, b):
    """
    Combine selectors in a single instance of type t.
    """
    if isinstance(a, t):
        a = a.conditions
    else:
        a = [a]
    if isinstance(b, t):
        b = b.conditions
    else:
        b = [b]
    return t(a + b)


class _Filter(_Unit, _Selector):
    """
    Wrapper class for all filters.
    """
    def __init__(self, unit):
        _Unit.__init__(self, unit)

    def __invert__(self):
        """
        Invert the filter (still act on the same event types, unary
        operator ~).
        """
        return self.invert()

    def __neg__(self):
        """
        Negate the filter (ignoring event types, unary operator -)
        """
        return self.negate()

    def invert(self):
        return _InvertedFilter(self, False)

    def negate(self):
        return _InvertedFilter(self, True)

    def build(self):
        return self

    def build_negated(self):
        return self.negate()


class _InvertedFilter(_Filter):
    """
    Inverted filter. keeps a reference to the original filter unit.
    """
    def __init__(self, filt, negate):
        self.filt = filt
        self.negate = negate
        _Filter.__init__(self, _mididings.InvertedFilter(filt.unit, negate))

    def __repr__(self):
        return _unitrepr.inverted_filter_to_string(self)


# the types we accept as part of a patch
_UNIT_TYPES = (_Unit, list, dict, _constants._EventType)
_SELECTOR_TYPES = (_Filter, _Selector, _constants._EventType)


@_arguments.accept([_UNIT_TYPES], add_varargs=True)
def Chain(units):
    """
    Units connected in series.
    """
    # handle a single list argument differently for backward compatibility
    if len(units) == 1 and type(units[0]) == list:
        units = units[0]

    return _Chain(units)


@_arguments.accept([_UNIT_TYPES], add_varargs=True,
                   kwargs={'remove_duplicates': (True, False, None)})
def Fork(units, **kwargs):
    """
    Units connected in parallel.
    """
    remove_duplicates = (kwargs['remove_duplicates']
                            if 'remove_duplicates' in kwargs else None)

    # handle a single list argument differently for backward compatibility
    if len(units) == 1 and type(units[0]) == list:
        units = units[0]

    return _Fork(units, remove_duplicates)


@_arguments.accept({
    _arguments.nullable(_arguments.reduce_bitmask([_constants._EventType])):
        _UNIT_TYPES
})
def Split(mapping):
    """
    Split(mapping)

    Split by event type.
    """
    return _Split(mapping)


@_arguments.accept([_SELECTOR_TYPES])
def And(conditions):
    """
    Conjunction of multiple filters.
    """
    return _AndSelector(conditions)


@_arguments.accept([_SELECTOR_TYPES])
def Or(conditions):
    """
    Disjunction of multiple filters.
    """
    return _OrSelector(conditions)


# for backward compatibility
AndSelector = And
OrSelector = Or


@_unitrepr.accept(_arguments.reduce_bitmask([_constants._EventType]),
                  add_varargs=True)
def Filter(types):
    """
    Filter(types, ...)

    Filter by event type. Multiple types can be given as bitmasks, lists, or
    separate parameters.
    """
    return _Filter(_mididings.TypeFilter(types))


@_unitrepr.store
def Pass():
    """
    Do nothing. This is sometimes useful/necessary as a placeholder, much like
    the ``pass`` statement in Python.
    """
    return _Unit(_mididings.Pass(True))


@_unitrepr.store
def Discard():
    """
    Stop processing the incoming event. Note that it is rarely neccessary to
    use this, as filters and splits already take care of removing unwanted
    events.
    """
    return _Unit(_mididings.Pass(False))
